// Copyright (c) 2023, Manfred Moitzi
// License: MIT License
// global constants
#define ABS_TOL 1e-12
#define REL_TOL 1e-9
#define M_TAU 6.283185307179586
// AutoCAD limits the degree of SPLINE to 11 or order = 12
#define MAX_SPLINE_ORDER 12
