# Copyright (c) 2014, Manfred Moitzi
# License: MIT License
# This module defines all message types for the internal messaging system:
# - ezdxf.messenger
# - DXFEntity.notify()


# Notify all entities to apply pending/temporary changes.
COMMIT_PENDING_CHANGES = 1